@@ -1,10 +1,15 @@
-from dataclasses import dataclass, field
+from pydantic import BaseModel, validator, ValidationError, Field, PrivateAttr
+from pydantic.dataclasses import dataclass
 from datetime import timezone
 from typing import Optional, Dict, List, Type, Any, Union, Tuple, Callable
-from fastbt.Meta import Broker
 import uuid
 import pendulum
+import sqlite3
+import logging
 from collections import Counter, defaultdict
+from omspy.base import Broker
+from copy import deepcopy
+from sqlite_utils import Database
 
 
 def get_option(spot: float, num: int = 0, step: float = 100.0) -> float:
@@ -23,12 +28,44 @@ def get_option(spot: float, num: int = 0, step: float = 100.0) -> float:
     v = round(spot / step)
     return v * (step + num)
 
-@dataclass
-class Order:
+
+def create_db(dbname: str = ":memory:") -> Union[Database, None]:
+    """
+    Create a sqlite3 database for the orders and return the connection
+    dbname
+        name of the database
+        default in-memory database
+    """
+    try:
+        con = sqlite3.connect(dbname)
+        with con:
+            con.execute(
+                """create table orders
+                           (symbol text, side text, quantity integer,
+                           id text primary key, parent_id text, timestamp text,
+                           order_type text, broker_timestamp text,
+                           exchange_timestamp text, order_id text,
+                           exchange_order_id text, price real,
+                           trigger_price real, average_price real,
+                           pending_quantity integer, filled_quantity integer,
+                           cancelled_quantity integer, disclosed_quantity integer,
+                           validity text, status text,
+                           expires_in integer, timezone text,
+                           client_id text, convert_to_market_after_expiry text,
+                           cancel_after_expiry text, retries integer, max_modifications integer,
+                           exchange text, tag string)"""
+            )
+            return Database(con)
+    except Exception as e:
+        logging.error(e)
+        return None
+
+
+class Order(BaseModel):
     symbol: str
     side: str
     quantity: int = 1
-    internal_id: Optional[str] = None
+    id: Optional[str] = None
     parent_id: Optional[str] = None
     timestamp: Optional[pendulum.DateTime] = None
     order_type: str = "MARKET"
@@ -51,23 +88,32 @@ class Order:
     convert_to_market_after_expiry: bool = False
     cancel_after_expiry: bool = True
     retries: int = 0
+    max_modifications: int = 10
     exchange: Optional[str] = None
     tag: Optional[str] = None
+    connection: Optional[Database] = None
+    _num_modifications: int = 0
+    _attrs: Tuple[str] = (
+        "exchange_timestamp",
+        "exchange_order_id",
+        "status",
+        "filled_quantity",
+        "pending_quantity",
+        "disclosed_quantity",
+        "average_price",
+    )
 
-    def __post_init__(self) -> None:
-        self.internal_id = uuid.uuid4().hex
+    class Config:
+        underscore_attrs_are_private = True
+        arbitrary_types_allowed = True
+
+    def __init__(self, **data) -> None:
+        super().__init__(**data)
+        if not (self.id):
+            self.id = uuid.uuid4().hex
         tz = self.timezone
         self.timestamp = pendulum.now(tz=tz)
         self.pending_quantity = self.quantity
-        self._attrs: List[str] = [
-            "exchange_timestamp",
-            "exchange_order_id",
-            "status",
-            "filled_quantity",
-            "pending_quantity",
-            "disclosed_quantity",
-            "average_price",
-        ]
         if self.expires_in == 0:
             self.expires_in = (
                 pendulum.today(tz=tz).end_of("day") - pendulum.now(tz=tz)
@@ -75,6 +121,12 @@ class Order:
         else:
             self.expires_in = abs(self.expires_in)
 
+    @validator("quantity", always=True)
+    def quantity_not_negative(cls, v):
+        if v < 0:
+            raise ValueError("quantity must be positive")
+        return v
+
     @property
     def is_complete(self) -> bool:
         if self.quantity == self.filled_quantity:
@@ -116,7 +168,7 @@ class Order:
     def has_parent(self) -> bool:
         return True if self.parent_id else False
 
-    def update(self, data: Dict[str, Any]) -> bool:
+    def update(self, data: Dict[str, Any], save: bool = True) -> bool:
         """
         Update order based on information received from broker
         data
@@ -132,11 +184,13 @@ class Order:
                 val = data.get(att)
                 if val:
                     setattr(self, att, val)
+            if self.connection and save:
+                self.save_to_db()
             return True
         else:
             return False
 
-    def execute(self, broker: Broker, **kwargs) -> Optional[str]:
+    def execute(self, broker: Type[Broker], **kwargs) -> Optional[str]:
         """
         Execute an order on a broker, place a new order
         kwargs
@@ -160,6 +214,8 @@ class Order:
             order_args.update(dct)
             order_id = broker.order_place(**order_args)
             self.order_id = order_id
+            if self.connection:
+                self.save_to_db()
             return order_id
         else:
             return self.order_id
@@ -177,8 +233,12 @@ class Order:
             "disclosed_quantity": self.disclosed_quantity,
         }
         dct = {k: v for k, v in kwargs.items() if k not in order_args.keys()}
-        order_args.update(dct)
-        broker.order_modify(**order_args)
+        order_args.update(**kwargs)
+        if self._num_modifications < self.max_modifications:
+            broker.order_modify(**order_args)
+            self._num_modifications += 1
+        else:
+            logging.info(f"Maximum number of modifications exceeded")
 
     def cancel(self, broker: Broker):
         """
@@ -186,19 +246,37 @@ class Order:
         """
         broker.order_cancel(order_id=self.order_id)
 
-@dataclass
-class CompoundOrder:
-    broker: Type[Broker]
-    internal_id: Optional[str] = None
+    def save_to_db(self) -> bool:
+        """
+        save or update the order to db
+        """
+        if self.connection:
+            values = self.dict(exclude={"connection"})
+            self.connection["orders"].upsert(values, pk="id")
+            return True
+        else:
+            logging.info("No valid database connection")
+            return False
 
-    def __post_init__(self) -> None:
-        self.internal_id = uuid.uuid4().hex
-        self._ltp: defaultdict = defaultdict()
-        self._orders: List[Order] = []
 
-    @property
-    def orders(self) -> List[Order]:
-        return self._orders
+class CompoundOrder(BaseModel):
+    broker: Any
+    id: Optional[str] = None
+    ltp: defaultdict = Field(default_factory=defaultdict)
+    orders: List[Order] = Field(default_factory=list)
+    connection: Optional[Database] = None
+    order_args: Optional[Dict] = None
+
+    class Config:
+        underscore_attrs_are_private = True
+        arbitrary_types_allowed = True
+
+    def __init__(self, **data) -> None:
+        super().__init__(**data)
+        if not (self.id):
+            self.id = uuid.uuid4().hex
+        if self.order_args is None:
+            self.order_args = {}
 
     @property
     def count(self) -> int:
@@ -207,10 +285,6 @@ class CompoundOrder:
         """
         return len(self.orders)
 
-    @property
-    def ltp(self) -> defaultdict:
-        return self._ltp
-
     @property
     def positions(self) -> Counter:
         """
@@ -227,10 +301,13 @@ class CompoundOrder:
         return c
 
     def add_order(self, **kwargs) -> Optional[str]:
-        kwargs["parent_id"] = self.internal_id
+        kwargs["parent_id"] = self.id
+        if not (kwargs.get("connection")):
+            kwargs["connection"] = self.connection
         order = Order(**kwargs)
-        self._orders.append(order)
-        return order.internal_id
+        order.save_to_db()
+        self.orders.append(order)
+        return order.id
 
     def _average_price(self, side: str = "buy") -> Dict[str, float]:
         """
@@ -274,10 +351,10 @@ class CompoundOrder:
         returns a dictionary with order_id and update status as boolean
         """
         dct: Dict[str, bool] = {}
-        for order in self.orders:
+        for order in self.pending_orders:
             order_id = str(order.order_id)
             status = order.status
-            if (order_id in data) and (status != "COMPLETE"):
+            if order_id in data:
                 d = data.get(order_id)
                 if d:
                     order.update(d)
@@ -324,7 +401,7 @@ class CompoundOrder:
         orders placed
         """
         for symbol, ltp in last_price.items():
-            self._ltp[symbol] = ltp
+            self.ltp[symbol] = ltp
         return self.ltp
 
     @property
@@ -358,9 +435,11 @@ class CompoundOrder:
     def total_mtm(self) -> float:
         return sum(self.mtm.values())
 
-    def execute_all(self):
+    def execute_all(self, **kwargs):
         for order in self.orders:
-            order.execute(broker=self.broker)
+            order_args = deepcopy(self.order_args)
+            order_args.update(kwargs)
+            order.execute(broker=self.broker, **order_args)
 
     def check_flags(self):
         """
@@ -382,204 +461,22 @@ class CompoundOrder:
     def pending_orders(self) -> List[Order]:
         return [order for order in self.orders if order.is_pending]
 
-
-    def __init__(
-        self,
-        symbol: str,
-        side: str,
-        trigger_price: float,
-        price: float = 0.0,
-        quantity: int = 1,
-        order_type="MARKET",
-        disclosed_quantity: int = 0,
-        order_args: Optional[Dict] = None,
-        **kwargs,
-    ):
-        super(StopOrder, self).__init__(**kwargs)
-        side2 = "sell" if side.lower() == "buy" else "buy"
-        self.add_order(
-            symbol=symbol,
-            side=side,
-            price=price,
-            quantity=quantity,
-            order_type=order_type,
-            disclosed_quantity=disclosed_quantity,
-        )
-        self.add_order(
-            symbol=symbol,
-            side=side2,
-            price=0,
-            trigger_price=trigger_price,
-            quantity=quantity,
-            order_type="SL-M",
-            disclosed_quantity=disclosed_quantity,
-        )
-
-class OptionStrategy:
-    """
-    Option Strategy is a list of compound orders
-    """
-
-    def __init__(self, broker: Type[Broker], profit=1e100, loss=-1e100) -> None:
-        self._orders: List[CompoundOrder] = []
-        self._broker: Type[Broker] = broker
-        self._ltp: defaultdict = defaultdict()
-        self.profit: float = profit
-        self.loss: float = loss
-
-    @property
-    def broker(self) -> Type[Broker]:
-        return self._broker
-
-    @property
-    def orders(self) -> List[CompoundOrder]:
-        return self._orders
-
-    def add_order(self, order: CompoundOrder) -> None:
+    def add(self, order: Order) -> Optional[str]:
         """
-        Add a compound order
-        broker is overriden
+        Add an order to the existing compound order
         """
-        order.broker = self.broker
-        self._orders.append(order)
+        order.parent_id = self.id
+        if not (order.connection):
+            order.connection = self.connection
+        order.save_to_db()
+        self.orders.append(order)
+        return order.id
 
-    @property
-    def all_orders(self) -> List[Order]:
+    def save(self):
         """
-        Get the list of all orders
+        Save all orders to database
         """
-        orders = []
-        for order in self.orders:
-            orders.extend(order.orders)
-        return orders
-
-    def update_ltp(self, last_price: Dict[str, float]) -> List[Any]:
-        """
-        Update ltp for the given symbols
-        last_price
-            dictionary with symbol as key and last price as value
-        """
-        return self._call("update_ltp", last_price=last_price)
-
-    def _call(self, attribute: str, **kwargs) -> List[Any]:
-        """
-        Call the given method or property on each of the compound orders
-        attribute
-            property or method
-        kwargs
-            keyword arguments to be called in case of a method
-        returns a list of the return values
-        Note
-        -----
-        1) An attribtute is considered to be a method if callable returns True
-        """
-        responses = []
-        for order in self.orders:
-            attr = getattr(order, attribute, None)
-            if callable(attr):
-                responses.append(attr(**kwargs))
-            else:
-                responses.append(attr)
-        return responses
-
-    def update_orders(self, data: Dict[str, Dict[str, Any]]) -> List[Any]:
-        """
-        Update all orders
-        data
-            data as dictionary with key as broker order_id
-        returns a dictionary with order_id and update status as boolean
-        for all compound orders
-        """
-        return self._call("update_orders", data=data)
-
-    def execute_all(self) -> List[Any]:
-        """
-        Execute all orders in all compound orders
-        """
-        return self._call("execute_all")
-
-    @property
-    def total_mtm(self) -> float:
-        """
-        Returns the total mtm for this strategy
-        """
-        mtm = self._call("total_mtm")
-        return sum([x for x in mtm if x is not None])
-
-    @property
-    def positions(self) -> Counter:
-        """
-        Return the combined positions for this strategy
-        """
-        c: Counter = Counter()
-        positions = self._call("positions")
-        for position in positions:
-            c.update(position)
-        return c
-
-    @property
-    def is_profit_hit(self) -> bool:
-        return True if self.total_mtm > self.profit else False
-
-    @property
-    def is_loss_hit(self) -> bool:
-        return True if self.total_mtm < self.loss else False
-
-    @property
-    def can_exit_strategy(self) -> bool:
-        """
-        Check whether we can exit from the strategy
-        We can exit from the strategy if either of the following
-        conditions is met
-        1) Profit is hit
-        2) Loss is hit
-        """
-        if self.is_profit_hit:
-            return True
-        elif self.is_loss_hit:
-            return True
-        else:
-            return False
-
-    """
-    Trailing stop order
-    """
-    def __init__(self, trail_by:Tuple[float, float],**kwargs):
-        self.trail_big:float = trail_by[0]
-        self.trail_small:float = trail_by[-1]
-        super(TrailingStopOrder, self).__init__(**kwargs)
-        self._maxmtm:float = 0
-        self._stop:float = kwargs.get('trigger_price', 0)
-        self.initial_stop = self._stop
-        self.symbol:str = kwargs.get('symbol')
-        self.quantity:int = kwargs.get('quantity',1)
-
-    @property
-    def stop(self):
-        return self._stop
-
-    @property
-    def maxmtm(self):
-        return self._maxmtm
-    
-    def _update_maxmtm(self):
-        self._maxmtm = max(self.total_mtm, self._maxmtm)
-
-    def _update_stop(self):
-        mtm_per_unit = self.maxmtm/self.quantity
-        multiplier = self.trail_small/self.trail_big
-        self._stop = self.initial_stop + (mtm_per_unit*multiplier)
-        
-    def watch(self):
-        self._update_maxmtm()
-        self._update_stop()
-        ltp = self.ltp.get(self.symbol)
-        if ltp:
-            #TODO: Implement for sell also
-            if ltp < self.stop:
-                order = self.orders[-1]
-                order.order_type = "MARKET"
-                order.modify(broker=self.broker)
-
-
-
+        if self.count > 0:
+            for order in self.orders:
+                order.save_to_db()
+        pass
